@@ -14,7 +14,9 @@ There are several units that are specifically designed to store meta variables:
   sub-pipeline inside a meta variable; more on this later.
 - The `refinery.cm` unit is a catch-all helper to generate common metadata such as size, frame
   index, hashes, entropy, etcetera.
-- The unit `refinery.wm` can be used to remove all metadata from a chunk.
+- The unit `refinery.mvc` can be used to clear local variables.
+- By default, variables exist only throughout the `refinery.lib.frame` that they are defined in.
+  The unit `refinery.mvg` can be used to propagate variables to parent frames.
 - The `refinery.struct` parses structured data from the beginning of a chunk into meta variables.
 - You can use named capture groups in regular expressions when using the `refinery.rex` unit, and
   these matches will be stored under their name as a meta variable in each output chunk.
@@ -92,9 +94,14 @@ The `refinery.push` and `refinery.pop` units can be used to extract sub-pipeline
 example, the following command extracts the files from a password-protected attachment of an email
 message by first extracting the password from the email message body:
 
-    $ emit phish.eml | push [[
-    >   | xtmail body.txt | rex -I password:\s*(\w+) {1} | pop password ]
-    >   | xtmail evil.zip | xtzip -p var:password | dump {path} ]
+    $ emit phish.eml [                      |
+    >     push [                            |
+    >         xtmail body.txt               |
+    >         rex -I password:\s*(\w+) {1}  |
+    >         pop password ]                |
+    >       xt *.zip                        |
+    >       xt *.exe -p var:password        |
+    >       dump extracted/{path} ]
 
 The `refinery.push` unit emits two copies of the input data, and the second copy has been moved out
 of scope (it is not visible). The first `refinery.xtmail` unit extracts the `body.txt` part and we
@@ -109,6 +116,7 @@ import abc
 import contextlib
 import string
 import codecs
+import itertools
 
 from io import StringIO
 from urllib.parse import quote_from_bytes, unquote_to_bytes
@@ -120,10 +128,12 @@ from refinery.lib.tools import isbuffer, entropy, index_of_coincidence
 
 if TYPE_CHECKING:
     from typing import Protocol
+    from refinery.lib.frame import Chunk
 
     class _Derivation(Protocol):
         costly: bool
         name: str
+        wrap: type
 
         def __call__(self, object: LazyMetaOracle) -> Union[str, int, float]:
             ...
@@ -221,10 +231,15 @@ class ByteStringWrapper(bytearray, CustomStringRepresentation):
         except AttributeError:
             pass
         try:
-            if self and not any(self[1::2]):
-                representation = self.decode('utf-16le')
-                prefix = 'u'
+            if not self or any(self[1::2]):
+                prefix = None
             else:
+                try:
+                    representation = self.decode('utf-16le')
+                    prefix = 'u'
+                except UnicodeDecodeError:
+                    prefix = None
+            if prefix is None:
                 representation = self.string
                 prefix = self._CODECS[self.codec]
         except AttributeError:
@@ -268,6 +283,8 @@ def is_valid_variable_name(name: str) -> bool:
         return False
     if not name.isidentifier():
         return False
+    if name == 'index':
+        return False
     return True
 
 
@@ -333,7 +350,7 @@ class _NoDerivationAvailable(Exception):
 
 class _LazyMetaMeta(type):
     def __new__(cls, name: str, bases, namespace: dict):
-        derivations: dict = namespace['DERIVATION_MAP']
+        derivations: dict = namespace['derivations']
         for obj in namespace.values():
             try:
                 obj: _Derivation
@@ -343,27 +360,23 @@ class _LazyMetaMeta(type):
         return type.__new__(cls, name, bases, namespace)
 
 
-def _derivation(name, costly: bool = False) -> Callable[[_Derivation], _Derivation]:
+def _derivation(name, costly: bool = False, wrap: type = ByteStringWrapper) -> Callable[[_Derivation], _Derivation]:
     def decorator(method: _Derivation) -> _Derivation:
         method.name = name
         method.costly = costly
+        method.wrap = wrap
         return method
     return decorator
 
 
-class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
+class LazyMetaOracle(metaclass=_LazyMetaMeta):
     """
     A dictionary that can be queried lazily for all potential options of the common meta variable
     unit. For example, a SHA-256 hash is computed only as soon as the oracle is accessed at the
     key `'sha256'`.
     """
 
-    CUSTOM_TYPE_MAP = {
-        'entropy' : Percentage,
-        'size'    : SizeInt,
-    }
-
-    DERIVATION_MAP: Dict[str, _Derivation] = {}
+    derivations: Dict[str, _Derivation] = {}
     """
     A dictionary mapping the names of common properties to anonymous functions that compute their
     corresponding value on a chunk of binary input data.
@@ -373,42 +386,154 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
     chunk: ByteString
     cache: Dict[str, Union[str, int, float]]
 
-    def __init__(self, chunk: ByteString, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, chunk: ByteString, scope: Optional[int] = 1, seed: Optional[Dict[str, list]] = None):
         self.ghost = False
         self.chunk = chunk
         self.cache = {}
-        self.fix()
+        self.scope = scope
+        self.tempval = {}
+        self.current = {}
+        self.updated = {}
+        self.rescope = {}
+        if seed is not None:
+            for key, stack in seed.items():
+                if not isinstance(stack, list):
+                    raise TypeError(R'Incorrect type in variable scope history.')
+                if len(stack) != scope:
+                    raise ValueError(F'History item had length {len(stack)}, but scope was specified as {scope}.')
+                for value in reversed(stack):
+                    if value is not None:
+                        self.current[key] = self.autowrap(key, value)
+                        self.updated[key] = False
+                        break
+                else:
+                    raise ValueError(R'History item was all None.')
+            self.history = seed
+        else:
+            self.history = {}
 
-    def update(self, *args, **kwargs):
-        super().update(*args, **kwargs)
-        self.fix()
-
-    def fix(self):
-        for key, value in self.items():
-            try:
-                ctype = self.CUSTOM_TYPE_MAP[key]
-            except KeyError:
-                if isinstance(value, ByteStringWrapper):
-                    continue
-                with contextlib.suppress(TypeError):
-                    self[key] = ByteStringWrapper(value)
-            else:
-                if not isinstance(value, ctype) and issubclass(ctype, type(value)):
-                    self[key] = ctype(value)
-        return self
+    def update(self, other: Union[dict, LazyMetaOracle]):
+        if isinstance(other, LazyMetaOracle):
+            self.current.update(other.current)
+            self.updated.update(other.updated)
+            self.tempval.update(other.tempval)
+            self.rescope.update(other.rescope)
+            self.history = other.history
+            return
+        for key, value in other.items():
+            self[key] = value
 
     def update_index(self, index: int):
         self['index'] = index
 
-    def serializable(self) -> dict:
-        return self
+    def inherit(self, parent: LazyMetaOracle):
+        if self.history and self.history is not parent.history:
+            raise RuntimeError
+        self.history = parent.history
+        self.scope = parent.scope
+        for key in parent.keys():
+            try:
+                derivation = self.derivations[key]
+            except KeyError:
+                pass
+            else:
+                if derivation.costly and len(self.chunk) >= 0x1000:
+                    continue
+                self[key] = derivation.wrap(derivation(self))
+
+    def set_scope(self, key: str, scope: int):
+        current = self.scope
+        scope = max(1, scope)
+        if key not in self.current:
+            raise KeyError(key)
+        if scope > current:
+            raise ValueError(F'Attempt to increase scope level of variable {key} to {scope}, it is currently at {self.scope}.')
+        if scope == current:
+            return
+        self.rescope[key] = scope
+
+    def get_scope(self, key: str):
+        value = self.current[key]
+        scope = self.scope
+        try:
+            stack = self.history[key]
+        except KeyError:
+            return scope
+        for k, v in enumerate(reversed(stack)):
+            if v is not None:
+                if v == value:
+                    return scope - k
+                break
+        return scope
+
+    def serialize(self, scope: int):
+        if not scope:
+            return {}
+        current_scope = self.scope
+        vanishing_variables = set()
+        serializable = {
+            key: list(stack) for key, stack in self.history.items()
+        }
+        if scope > current_scope:
+            padding = scope - current_scope
+            for key, stack in serializable.items():
+                stack.extend(itertools.repeat(None, padding))
+        if scope < current_scope:
+            for key, stack in serializable.items():
+                del stack[scope:]
+                if all(v is None for v in stack):
+                    vanishing_variables.add(key)
+        for key, value in self.current.items():
+            if value is None:
+                raise RuntimeError(F'Meta variable "{key}" was set to None.')
+            try:
+                spot = self.rescope[key]
+            except KeyError:
+                spot = current_scope
+            if spot == current_scope and not self.updated[key]:
+                continue
+            if spot > scope:
+                continue
+            if spot < 0:
+                raise RuntimeError('computed a negative spot for variable placement')
+            last_scope = None
+            last_value = None
+            try:
+                stack = serializable[key]
+            except KeyError:
+                serializable[key] = stack = [None] * scope
+            else:
+                vanishing_variables.discard(key)
+                for k, v in enumerate(reversed(stack)):
+                    if v is not None:
+                        last_scope = scope - k
+                        last_value = v
+                        break
+            if value == last_value:
+                spot = max(scope, last_scope)
+            stack[spot - 1] = value
+            unpadded_range = min(scope, current_scope)
+            if spot < unpadded_range:
+                stack[spot:unpadded_range] = itertools.repeat(None, unpadded_range - spot)
+        for key in vanishing_variables:
+            del serializable[key]
+        return serializable
 
     def items(self):
-        for key, value in super().items():
-            if not is_valid_variable_name(key):
-                continue
-            yield key, value
+        yield from self.tempval.items()
+        yield from self.current.items()
+
+    def keys(self):
+        yield from self.tempval.keys()
+        yield from self.current.keys()
+
+    def variable_names(self):
+        yield from self.current.keys()
+
+    def values(self):
+        return (v for _, v in self.items())
+
+    __iter__ = keys
 
     def format_str(
         self,
@@ -492,7 +617,6 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
         if fixup:
             for (store, it) in (
                 (args, enumerate(args)),
-                (self, self.items()),
                 (symb, symb.items()),
             ):
                 for key, value in it:
@@ -607,23 +731,108 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
             return stream.getvalue()
 
     def knows(self, key):
-        return super().__contains__(key) or self.cache.__contains__(key)
+        return (
+            key in self.current or # noqa
+            key in self.tempval or # noqa
+            key in self.cache
+        )
 
     def __contains__(self, key):
-        return super().__contains__(key) or key in self.DERIVATION_MAP
+        return (
+            key in self.current or # noqa
+            key in self.tempval or # noqa
+            key in self.derivations
+        )
+
+    def __len__(self):
+        return len(self.current) + len(self.tempval)
+
+    def autowrap(self, key, value):
+        try:
+            wrap = self.derivations[key].wrap
+        except KeyError:
+            wrap = ByteStringWrapper
+        if not isinstance(value, wrap):
+            with contextlib.suppress(TypeError):
+                value = ByteStringWrapper(value)
+        return value
+
+    def __setitem__(self, key, value):
+        new = self.autowrap(key, value)
+        if not is_valid_variable_name(key):
+            self.tempval[key] = new
+            return
+        self.current[key] = new
+        try:
+            stack = self.history[key]
+        except KeyError:
+            self.updated[key] = True
+        else:
+            for old in reversed(stack):
+                if old is not None:
+                    break
+            self.updated[key] = (old != new)
+
+    class nodefault:
+        pass
+
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
+
+    def pop(self, key, default=nodefault):
+        try:
+            value = self.current.pop(key)
+        except KeyError:
+            try:
+                return self.tempval.pop(key)
+            except KeyError:
+                pass
+            if default is self.nodefault:
+                raise
+            return default
+        else:
+            return value
 
     def __getitem__(self, key):
-        item = super().__getitem__(key)
-        if isinstance(item, str):
-            return item.encode('utf8')
-        return item
+        try:
+            value = self.current[key]
+        except KeyError:
+            try:
+                return self.tempval[key]
+            except KeyError:
+                pass
+            return self.__missing__(key)
+        if isinstance(value, str):
+            value = value.encode('utf8')
+        return value
+
+    def discard(self, key):
+        try:
+            del self.current[key]
+        except KeyError:
+            try:
+                del self.tempval[key]
+            except KeyError:
+                pass
+        try:
+            stack = self.history[key]
+        except KeyError:
+            pass
+        else:
+            if stack:
+                stack[-1] = None
+
+    __delitem__ = discard
 
     def __getattr__(self, key):
-        if not super().__contains__(key):
-            deduction = self.DERIVATION_MAP.get(key)
+        if key not in self.current:
+            deduction = self.derivations.get(key)
             if deduction is None:
                 raise AttributeError(key)
-            return deduction(self)
+            return deduction.wrap(deduction(self))
         else:
             return self[key]
 
@@ -632,11 +841,11 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
             return self.cache[key]
         except KeyError:
             pass
-        deduction = self.DERIVATION_MAP.get(key)
+        deduction = self.derivations.get(key)
         if deduction is None:
             raise KeyError(F'The meta variable {key} is unknown.')
         try:
-            value = deduction(self)
+            value = deduction.wrap(deduction(self))
         except _NoDerivationAvailable:
             raise KeyError(F'unable to derive the {key} property here, you have to use the cm unit.')
         else:
@@ -646,12 +855,6 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
     def derive(self, key):
         self[key] = self[key]
 
-    def discard(self, key):
-        try:
-            dict.__delitem__(self, key)
-        except KeyError:
-            pass
-
     @_derivation('mime')
     def _derive_mime(self):
         from refinery.lib.mime import get_cached_file_magic_info
@@ -667,54 +870,55 @@ class LazyMetaOracle(dict, metaclass=_LazyMetaMeta):
         from refinery.lib.mime import get_cached_file_magic_info
         return get_cached_file_magic_info(self.chunk).description
 
-    @_derivation('size')
+    @_derivation('size', wrap=SizeInt)
     def _derive_size(self):
-        return SizeInt(len(self.chunk))
+        return len(self.chunk)
 
-    @_derivation('entropy', True)
+    @_derivation('entropy', True, Percentage)
     def _derive_entropy(self):
-        return Percentage(entropy(self.chunk))
+        return entropy(self.chunk)
 
-    @_derivation('ic', True)
+    @_derivation('ic', True, Percentage)
     def _derive_ic(self):
-        return Percentage(index_of_coincidence(self.chunk))
+        return index_of_coincidence(self.chunk)
 
     @_derivation('crc32')
     def _derive_crc32(self):
         import zlib
-        return ByteStringWrapper((zlib.crc32(self.chunk) & 0xFFFFFFFF).to_bytes(4, 'big').hex())
+        return (zlib.crc32(self.chunk) & 0xFFFFFFFF).to_bytes(4, 'big').hex()
 
     @_derivation('sha1', True)
     def _derive_sha1(self):
         import hashlib
-        return ByteStringWrapper(hashlib.sha1(self.chunk).hexdigest())
+        return hashlib.sha1(self.chunk).hexdigest()
 
     @_derivation('sha256', True)
     def _derive_sha256(self):
         import hashlib
-        return ByteStringWrapper(hashlib.sha256(self.chunk).hexdigest())
+        return hashlib.sha256(self.chunk).hexdigest()
 
     @_derivation('sha512', True)
     def _derive_sha512(self):
         import hashlib
-        return ByteStringWrapper(hashlib.sha512(self.chunk).hexdigest())
+        return hashlib.sha512(self.chunk).hexdigest()
 
     @_derivation('md5', True)
     def _derive_md5(self):
         import hashlib
-        return ByteStringWrapper(hashlib.md5(self.chunk).hexdigest())
+        return hashlib.md5(self.chunk).hexdigest()
 
 
-def metavars(chunk, ghost: bool = False) -> LazyMetaOracle:
+def metavars(chunk: Union[Chunk, ByteString]) -> LazyMetaOracle:
     """
     This method is the main function used by refinery units to get the meta variable dictionary
     of an input chunk. This dictionary is wrapped using the `refinery.lib.meta.LazyMetaOracleFactory`
     so that access to common variables is always possible.
     """
-    meta: Union[dict, LazyMetaOracle] = getattr(chunk, 'meta', {})
-    if not isinstance(meta, LazyMetaOracle):
-        if not isinstance(meta, dict):
-            raise TypeError('Invalid meta type.')
-        meta = LazyMetaOracle(chunk, meta)
-    meta.ghost = ghost
+    try:
+        meta = chunk.meta
+    except AttributeError:
+        meta = LazyMetaOracle(chunk)
+    else:
+        if not isinstance(meta, LazyMetaOracle):
+            raise TypeError(F'Invalid meta variable dictionary on chunk: {meta!r}')
     return meta
