@@ -676,9 +676,11 @@ def UnitProcessorBoilerplate(operation: ProcType[ByteString]) -> ProcType[Chunk]
         if not isinstance(data, ChunkType):
             data = ChunkType(data)
         result = operation(self, data)
-        if not inspect.isgenerator(result):
-            return self.labelled(result)
-        return (self.labelled(r) for r in result)
+        if isinstance(result, Chunk):
+            return result
+        elif not inspect.isgenerator(result):
+            return Chunk(result)
+        return (Chunk.Wrap(r) for r in result)
     return wrapped
 
 
@@ -851,12 +853,7 @@ class Executable(ABCMeta):
             cls.__init__ = cls__init__
 
         if not abstract and sys.modules[cls.__module__].__name__ == '__main__':
-            if Executable.Entry:
-                cls.logger.warning(cls._output(
-                    F'not executing this unit because the following unit was '
-                    F'already executed: {Executable.Entry}'
-                ))
-            else:
+            if not Executable.Entry:
                 Executable.Entry = cls.name
                 cls.run()
 
@@ -1236,7 +1233,7 @@ class Unit(UnitBase, abstract=True):
             import traceback
             traceback.print_exc(file=sys.stderr)
         if abort_execution:
-            raise RefineryCriticalException
+            raise RefineryCriticalException(str(exception))
 
     def __next__(self) -> Chunk:
         if not self._chunks:
@@ -1474,7 +1471,11 @@ class Unit(UnitBase, abstract=True):
                 return (self | stdout).getvalue()
         elif callable(stream):
             with MemoryFile() as stdout:
-                return stream((self | stdout).getvalue())
+                self | stdout
+                out = stdout.getbuffer()
+                if isinstance(stream, type) and isinstance(out, stream):
+                    return out
+                return stream(out)
 
         if not stream.writable():
             raise ValueError('target stream is not writable')
